@@ -20,6 +20,8 @@ parameter expansion.
 #include <unistd.h>
 #include <signal.h>
 
+#include <assert.h>
+
 #ifdef SunOS
 #include <procfs.h>
 #endif
@@ -29,10 +31,10 @@ parameter expansion.
 #include "common.h"
 #include "wutil.h"
 #include "env.h"
+#include "proc.h"
 #include "parser.h"
 #include "expand.h"
 #include "wildcard.h"
-#include "proc.h"
 #include "exec.h"
 #include "tokenizer.h"
 #include "complete.h"
@@ -57,13 +59,18 @@ parameter expansion.
 */
 #define COMPLETE_JOB_DESC_VAL COMPLETE_SEP_STR, L"Job: "
 
-
+/**
+   Description for the shells own pid
+*/
 #define COMPLETE_SELF_DESC COMPLETE_SEP_STR L"Shell process"
 
 /**
-   Separate subtokens in a token with this character. 
+   Description for the shells own pid
 */
-#define INTERNAL_SEPARATOR 0xfffffff0
+#define COMPLETE_LAST_DESC COMPLETE_SEP_STR L"Last background job"
+
+#define COMPLETE_VAR_DESC L"Variable name is zero characters long."
+#define COMPLETE_VAR2_DESC L"Variable name is zero characters long.  Did you mean{$VARIABLE}? To learn about variable expansion in fish, type 'help expand-variable'."
 
 /**
    String in process expansion denoting ourself
@@ -71,13 +78,60 @@ parameter expansion.
 #define SELF_STR L"self"
 
 /**
-   Return the environment variable value for the string starting at in 
+   String in process expansion denoting last background job
+*/
+#define LAST_STR L"last"
+
+/**
+   Characters which make a string unclean if they are the first
+   character of the string. See \c is_clean().
+*/
+#define UNCLEAN_FIRST L"~%"
+/**
+   Unclean characters. See \c is_clean().
+*/
+#define UNCLEAN L"$*?\\\"'({})"
+
+/**
+   Test if the specified argument is clean, i.e. it does not contain
+   any tokens which need to be expanded or otherwise altered. Clean
+   strings can be passed through expand_string and expand_one without
+   changing them. About 90% of all strings are clean, so skipping
+   expansion on them actually does save a small amount of time, since
+   it avoids multiple memory allocations during the expansion process.
+*/
+static int is_clean( const wchar_t *in )
+{
+	
+	const wchar_t * str = in;
+
+	/*
+	  Test characters that have a special meaning in the first character position
+	*/
+	if( wcschr( UNCLEAN_FIRST, *str ) )
+		return 0;
+	
+	/*
+	  Test characters that have a special meaning in any character position
+	*/
+	while( *str )
+	{
+		if( wcschr( UNCLEAN, *str ) )
+			return 0;
+		str++;
+	}
+	
+	return 1;
+}
+
+/**
+   Return the environment variable value for the string starting at \c in. 
 */
 static wchar_t *expand_var( wchar_t *in )
 {
 	if( !in )
 		return 0;
-	return (in[0] == VARIABLE_EXPAND )? env_get( expand_var(in+1) ) : env_get( in );
+	return env_get( in );
 }
 
 void expand_variable_array( const wchar_t *val, array_list_t *out )
@@ -87,11 +141,9 @@ void expand_variable_array( const wchar_t *val, array_list_t *out )
 		wchar_t *cpy = wcsdup( val );
 		wchar_t *pos, *start;
 		
-		if( (cpy==0) )
+		if( !cpy )
 		{
-/*		sb_append( sb_err, 
-  L"fish: Out of memory in expand_variable_array()\n" );*/
-			return;
+			die_mem();
 		}
 
 		for( start=pos=cpy; *pos; pos++ )
@@ -104,89 +156,18 @@ void expand_variable_array( const wchar_t *val, array_list_t *out )
 			}
 		}
 		al_push( out, wcsdup(start) );
-	
+		
 		free(cpy);	
 	}
 }
 
-wchar_t *expand_escape( wchar_t *in, 
+wchar_t *expand_escape( const wchar_t *in,
 						int escape_all )
 {
-	wchar_t *killme=in;	
-	wchar_t *out = malloc( sizeof(wchar_t)*(wcslen(in)*2 + 1));
-	if( out != 0 )
-	{
-		wchar_t *pos=out;
-		while( *in != 0 )
-		{
-			switch( *in )
-			{
-				case L'\t':
-					*(pos++) = L'\\';
-					*(pos++) = L't';					
-					break;
-					
-				case L'\n':
-					*(pos++) = L'\\';
-					*(pos++) = L'n';					
-					break;
-					
-				case L'\b':
-					*(pos++) = L'\\';
-					*(pos++) = L'b';					
-					break;
-					
-				case L'\r':
-					*(pos++) = L'\\';
-					*(pos++) = L'r';					
-					break;
-					
-				case L'\e':
-					*(pos++) = L'\\';
-					*(pos++) = L'e';					
-					break;
-					
-				case ARRAY_SEP:
-					*(pos++) = L' ';
-					break;
-					
-				case L'\\':
-				case L'&':
-				case L'$':
-				case L' ':
-				case L'#':
-				case L'^':
-				case L'<':
-				case L'>':
-				case L'@':
-				case L'(':
-				case L')':
-				case L'{':
-				case L'}':
-				case L'?':
-				case L'*':
-				case L'|':
-				case L';':
-				case L':':
-				case L'\'':
-				case L'\"':
-					if( escape_all )
-						*pos++ = L'\\';
-					*pos++ = *in;
-					break;
-
-				default:
-					*pos++ = *in;
-					break;
-			}
-			in++;
-		}
-		*pos = 0;
-		free(killme);		
-	}
-	return out;
+	return escape( in, escape_all );
 }
 
+
 /**
    Test if the specified string does not contain character which can
    not be used inside a quoted string.
@@ -225,7 +206,7 @@ wchar_t *expand_escape_variable( const wchar_t *in )
 	switch( al_get_count( &l) )
 	{
 		case 0:
-			sb_append( &buff, L"\'\'");
+			sb_append( &buff, L"''");
 			break;			
 
 		case 1:
@@ -235,14 +216,14 @@ wchar_t *expand_escape_variable( const wchar_t *in )
 			if( wcschr( el, L' ' ) && is_quotable( el ) )
 			{
 				sb_append2( &buff,
-							L"\'",
+							L"'",
 							el,
-							L"\'",
-							0 );
+							L"'",
+							(void *)0 );
 			}
 			else
 			{
-				wchar_t *val = expand_escape( wcsdup(el), 1 );								
+				wchar_t *val = expand_escape( el, 1 );								
 				sb_append( &buff, val );
 				free( val );
 			}
@@ -261,14 +242,14 @@ wchar_t *expand_escape_variable( const wchar_t *in )
 				if( is_quotable( el ) )
 				{
 					sb_append2( &buff,
-								L"\'",
+								L"'",
 								el,
-								L"\'",
-								0 );
+								L"'",
+								(void *)0 );
 				}
 				else
 				{
-					wchar_t *val = expand_escape( wcsdup(el), 1 );								
+					wchar_t *val = expand_escape( el, 1 );								
 					sb_append( &buff, val );
 					free( val );
 				}
@@ -313,14 +294,14 @@ static int iswnumeric( const wchar_t *n )
 */
 static int match_pid( const wchar_t *cmd, 
 					  const wchar_t *proc,
-					  int accept_incomplete )
+					  int flags )
 {
 	/* Test for direct match */
 	
 	if( wcsncmp( cmd, proc, wcslen( proc ) ) == 0 )
 		return 1;
 	
-	if( accept_incomplete )
+	if( flags & ACCEPT_INCOMPLETE )
 		return 0;
 
 	/* 
@@ -364,7 +345,8 @@ static int match_pid( const wchar_t *cmd,
    Searches for a job with the specified job id, or a job or process
    which has the string \c proc as a prefix of its commandline.
 
-   If accept_incomplete is true, the remaining string for any matches are inserted.
+   If accept_incomplete is true, the remaining string for any matches
+   are inserted.
 
    If accept_incomplete is false, any job matching the specified
    string is matched, and the job pgid is returned. If no job
@@ -374,7 +356,7 @@ static int match_pid( const wchar_t *cmd,
 */
 
 static int find_process( const wchar_t *proc, 
-						 int accept_incomplete, 
+						 int flags, 
 						 array_list_t *out )
 {
 	DIR *dir;
@@ -387,12 +369,16 @@ static int find_process( const wchar_t *proc,
 	wchar_t *result;	
 	
 	job_t *j;
-
-	if( iswnumeric(proc ) || (wcslen(proc)==0) )
+	
+	if( iswnumeric(proc) || (wcslen(proc)==0) )
 	{
+		/*
+		  This is a numeric job string, like '%2' 
+		*/
+
 		//	fwprintf( stderr, L"Numeric\n\n\n" );
 
-		if( accept_incomplete )
+		if( flags & ACCEPT_INCOMPLETE )
 		{
 			for( j=first_job; j != 0; j=j->next )
 			{
@@ -445,9 +431,9 @@ static int find_process( const wchar_t *proc,
 		
 		//	fwprintf( stderr, L"match '%ls' '%ls'\n\n\n", j->command, proc );
 		
-		if( match_pid( j->command, proc, accept_incomplete ) )
+		if( match_pid( j->command, proc, flags ) )
 		{
-			if( accept_incomplete )
+			if( flags & ACCEPT_INCOMPLETE )
 			{
 				wchar_t *res = wcsdupcat( j->command + wcslen(proc), 
 										  COMPLETE_JOB_DESC );
@@ -484,9 +470,9 @@ static int find_process( const wchar_t *proc,
 			
 			//	fwprintf( stderr, L"match '%ls' '%ls'\n\n\n", j->command, proc );
 			
-			if( match_pid( p->actual_cmd, proc, accept_incomplete ) )
+			if( match_pid( p->actual_cmd, proc, flags ) )
 			{
-				if( accept_incomplete )
+				if( flags & ACCEPT_INCOMPLETE )
 				{
 					wchar_t *res = wcsdupcat( p->actual_cmd + wcslen(proc), 
 											  COMPLETE_CHILD_PROCESS_DESC );
@@ -594,9 +580,9 @@ static int find_process( const wchar_t *proc,
 		
 		if( cmd != 0 )
 		{
-			if( match_pid( cmd, proc, accept_incomplete ) )
+			if( match_pid( cmd, proc, flags ) )
 			{			
-				if( accept_incomplete )
+				if( flags & ACCEPT_INCOMPLETE )
 				{
 					wchar_t *res = wcsdupcat( cmd + wcslen(proc),
 											  COMPLETE_PROCESS_DESC );
@@ -607,20 +593,20 @@ static int find_process( const wchar_t *proc,
 				{
 					wchar_t *res = str2wcs(name);
 					if( res )
-						al_push( out, str2wcs( name ) );
+						al_push( out, res );
 				}
 			}
 		}
 	}
-
+	
 	if( cmd != 0 )
 		free( cmd );
-
+	
 	free( pdir_name );
 	free( pfile_name );
-
+	
 	closedir( dir );	
-
+	
 	return 1;
 }
 
@@ -628,7 +614,7 @@ static int find_process( const wchar_t *proc,
    Process id expansion
 */
 static int expand_pid( wchar_t *in, 
-					   int accept_incomplete,
+					   int flags,
 					   array_list_t *out )
 {
 	if( *in != PROCESS_EXPAND )
@@ -637,13 +623,18 @@ static int expand_pid( wchar_t *in,
 		return 1;
 	}
 
-	if( accept_incomplete )
+	if( flags & ACCEPT_INCOMPLETE )
 	{
 		if( wcsncmp( in+1, SELF_STR, wcslen(in+1) )==0 )
 		{
 			wchar_t *res = wcsdupcat( SELF_STR+wcslen(in+1), COMPLETE_SELF_DESC );
 			al_push( out, res );			
 		}		
+		else if( wcsncmp( in+1, LAST_STR, wcslen(in+1) )==0 )
+		{
+			wchar_t *res = wcsdupcat( LAST_STR+wcslen(in+1), COMPLETE_LAST_DESC );
+			al_push( out, res );			
+		}		
 	}
 	else
 	{
@@ -656,19 +647,32 @@ static int expand_pid( wchar_t *in,
 			
 			return 1;			
 		}		
-	}
-	
+		if( wcscmp( (in+1), LAST_STR )==0 )
+		{
+			wchar_t *str;
+
+			if( proc_last_bg_pid > 0 )
+			{
+				str = malloc( sizeof(wchar_t)*32);
+				free(in);
+				swprintf( str, 32, L"%d", proc_last_bg_pid );
+				al_push( out, str );
+			}
+			
+			return 1;			
+		}		
+	}	
 	
 //	fwprintf( stderr, L"expand_pid() %ls\n", in );
 	int prev = al_get_count( out );
-	if( !find_process( in+1, accept_incomplete, out ) )
+	if( !find_process( in+1, flags, out ) )
 		return 0;
 	
 	if( prev == al_get_count( out ) )
 	{
 //		fwprintf( stderr, L"no match\n" );
 		
-		if( accept_incomplete )
+		if( flags & ACCEPT_INCOMPLETE )
 			free( in );
 		else
 		{
@@ -698,11 +702,10 @@ static int expand_variables( wchar_t *in, array_list_t *out )
 	int is_ok= 1;
 	int empty=0;
 	
-
 	for( i=wcslen(in)-1; (i>=0) && is_ok && !empty; i-- )
 	{
 		c = in[i];
-		if( c == VARIABLE_EXPAND )
+		if( ( c == VARIABLE_EXPAND ) || (c == VARIABLE_EXPAND_SINGLE ) )
 		{
 			int start_pos = i+1;
 			int stop_pos;
@@ -711,15 +714,8 @@ static int expand_variables( wchar_t *in, array_list_t *out )
 			wchar_t * var_val;
 			wchar_t * new_in;			
 			array_list_t l;
+			int is_single = (c==VARIABLE_EXPAND_SINGLE);
 			
-			
-
-//			fwprintf( stderr, L"Expand %ls\n", in );
-			
-									
-//			while (in[stop_pos]==VARIABLE_EXPAND)			 
-//				stop_pos++;
-
 			stop_pos = start_pos;
 
 			while( 1 )
@@ -736,107 +732,157 @@ static int expand_variables( wchar_t *in, array_list_t *out )
 /*			printf( "Stop for '%c'\n", in[stop_pos]);*/
 			
 			var_len = stop_pos - start_pos;
+
+			if( var_len == 0 )
+			{
+				if( in[stop_pos] == BRACKET_BEGIN )
+				{
+										
+					error( SYNTAX_ERROR,
+						   -1, COMPLETE_VAR_DESC
+						   COMPLETE_VAR2_DESC );
+				}
+				else
+				{
+					error( SYNTAX_ERROR,
+						   -1,
+						   COMPLETE_VAR_DESC);
+				}
+				
+				is_ok = 0;
+				break;
+			}
+			
 			
 			if( !(var_name = malloc( sizeof(wchar_t)*(var_len+1) )))
 			{
-				error( OOM, L"Out of memory", -1 );				
-				is_ok = 0;
+				die_mem();
 			}
-			else
-			{
-				wcsncpy( var_name, &in[start_pos], var_len );
-				var_name[var_len]='\0';
+			wcsncpy( var_name, &in[start_pos], var_len );
+			var_name[var_len]='\0';
 /*			printf( "Variable name is %s, len is %d\n", var_name, var_len );*/
-				wchar_t *var_val_orig = expand_var( var_name );
+			wchar_t *var_val_orig = expand_var( var_name );
 		
-				if( var_val_orig && (var_val = wcsdup( var_val_orig) ) )
-				{
-					int all_vars=1;
-					array_list_t idx;
-					al_init( &idx );
-					al_init( &l );
+			if( var_val_orig && (var_val = wcsdup( var_val_orig) ) )
+			{
+				int all_vars=1;
+				array_list_t idx;
+				al_init( &idx );
+				al_init( &l );
 					
-					if( in[stop_pos] == L'[' )
-					{						
-						wchar_t *end;
+				if( in[stop_pos] == L'[' )
+				{						
+					wchar_t *end;
 						
-						all_vars = 0;
+					all_vars = 0;
 		
-						stop_pos++;
-						while( 1 )
+					stop_pos++;
+					while( 1 )
+					{
+						int tmp;
+							
+						while( iswspace(in[stop_pos]) || (in[stop_pos]==INTERNAL_SEPARATOR))
+							stop_pos++;
+							
+							
+						if( in[stop_pos] == L']' )
 						{
-							int tmp;
-							
-							while( iswspace(in[stop_pos]) || (in[stop_pos]==INTERNAL_SEPARATOR))
-								stop_pos++;
-							
-							
-							if( in[stop_pos] == L']' )
-							{
-								stop_pos++;
-								break;
-							}
+							stop_pos++;
+							break;
+						}
 
-							errno=0;
-							tmp = wcstol( &in[stop_pos], &end, 10 );
-							if( ( errno ) || ( end == &in[stop_pos] ) )
+						errno=0;
+						tmp = wcstol( &in[stop_pos], &end, 10 );
+						if( ( errno ) || ( end == &in[stop_pos] ) )
+						{
+							error( SYNTAX_ERROR, 
+								   -1, 
+								   L"Expected integer or \']\'" );
+															   
+							is_ok = 0;
+							break;
+						}
+						al_push( &idx, (void *)tmp );
+						stop_pos = end-in;
+					}
+				}
+							
+				if( is_ok )
+				{				
+					expand_variable_array( var_val, &l );
+					if( !all_vars )
+					{	
+						int j;
+						for( j=0; j<al_get_count( &idx ); j++)
+						{
+							int tmp = (int)al_get( &idx, j );
+							if( tmp < 1 || tmp > al_get_count( &l ) )
 							{
 								error( SYNTAX_ERROR, 
-									   L"Expected integer or \']\'", 
-									   -1 );
-								is_ok = 0;
+									   -1,
+									   L"Array index out of bounds" );
+								is_ok=0;
+								al_truncate( &idx, j );
 								break;
-							}
-							al_push( &idx, (void *)tmp );
-							stop_pos = end-in;
-						}
-					}
-							
-					if( is_ok )
-					{				
-						expand_variable_array( var_val, &l );
-						if( !all_vars )
-						{	
-							int j;
-							for( j=0; j<al_get_count( &idx ); j++)
+							}				
+							else
 							{
-								int tmp = (int)al_get( &idx, j );
-								if( tmp < 1 || tmp > al_get_count( &l ) )
-								{
-									error( SYNTAX_ERROR, L"Array index out of bounds", -1 );
-									is_ok=0;
-									al_truncate( &idx, j );
-									break;
-								}				
-								else
-								{
-									/* Move string from list l to list idx */
-									al_set( &idx, j, al_get( &l, tmp-1 ) );
-									al_set( &l, tmp-1, 0 );
-								}
+								/* Move string from list l to list idx */
+								al_set( &idx, j, al_get( &l, tmp-1 ) );
+								al_set( &l, tmp-1, 0 );
 							}
-							/* Free remaining strings in list l and truncate it */
-							al_foreach( &l, (void (*)(const void *))&free );
-							al_truncate( &l, 0 );
-							/* Add items from list idx back to list l */							
-							al_push_all( &l, &idx );							
 						}
-						free( var_val );
-					}				
+						/* Free remaining strings in list l and truncate it */
+						al_foreach( &l, (void (*)(const void *))&free );
+						al_truncate( &l, 0 );
+						/* Add items from list idx back to list l */							
+						al_push_all( &l, &idx );							
+					}
+					free( var_val );
+				}				
+				
+				if( is_single )
+				{
+					string_buffer_t res;
+					sb_init( &res );
+					
+					in[i]=0;
+					
+					sb_append( &res, in );
+					sb_append_char( &res, INTERNAL_SEPARATOR );
 
+					for( j=0; j<al_get_count( &l); j++ )
+					{
+						wchar_t *next = (wchar_t *)al_get( &l, j );
+						
+						if( is_ok )
+						{						
+							if( j != 0 )
+								sb_append( &res, L" " );
+							sb_append( &res, next );
+						}
+						free( next );
+					}	
+					sb_append( &res, &in[stop_pos] );
+					is_ok &= expand_variables( wcsdup((wchar_t *)res.buff), out );
+					
+					sb_destroy( &res );
+					
+				}
+				else
+				{
 					for( j=0; j<al_get_count( &l); j++ )
 					{
 						wchar_t *next = (wchar_t *)al_get( &l, j );
 						
 						if( is_ok )
 						{
-							
+						
 							new_len = wcslen(in) - (stop_pos-start_pos+1) + wcslen( next) +2;
 							
 							if( !(new_in = malloc( sizeof(wchar_t)*new_len )))
 							{
-								error( OOM, L"Out of memory", -1 );				
-								is_ok = 0;
+								die_mem();
 							}
 							else
 							{
@@ -859,21 +905,54 @@ static int expand_variables( wchar_t *in, array_list_t *out )
 							}
 						}
 						free( next );
-					}					
-					al_destroy( &l );
-					al_destroy( &idx );
-					free(in);
-					free(var_name );
-					return is_ok;					
+					}	
+				}
+				
+				al_destroy( &l );
+				al_destroy( &idx );
+				free(in);
+				free(var_name );
+				return is_ok;					
+			}
+			else
+			{
+				/*
+				  Expand a non-existing variable
+				*/
+				if( c == VARIABLE_EXPAND )
+				{
+					/*
+					  Regular expansion, i.e. expand this argument to nothing
+					*/
+					empty = 1;
 				}
 				else
 				{
-					empty = 1;
+					/*
+					  Expansion to single argument. 
+					*/
+					string_buffer_t res;
+					sb_init( &res );
+					
+					in[i]=0;
+					
+					sb_append( &res, in );
+
+					sb_append( &res, &in[stop_pos] );
+					is_ok &= expand_variables( wcsdup((wchar_t *)res.buff), out );
+
+					sb_destroy( &res );
+					free(in);
+					free(var_name );
+					return is_ok;
 				}
 				
-				free(var_name );				
 			}
+				
+			free(var_name );				
+				
 		}
+		
 		prev_char = c;		
 	}
 	
@@ -892,7 +971,7 @@ static int expand_variables( wchar_t *in, array_list_t *out )
 /**
    Perform bracket expansion
 */
-static int expand_brackets( wchar_t *in, int accept_incomplete, array_list_t *out )
+static int expand_brackets( wchar_t *in, int flags, array_list_t *out )
 {
 	wchar_t *pos;
 	int syntax_error=0;
@@ -944,19 +1023,13 @@ static int expand_brackets( wchar_t *in, int accept_incomplete, array_list_t *ou
 	
 	if( bracket_count > 0 )
 	{
-		if( !accept_incomplete )
+		if( !(flags & ACCEPT_INCOMPLETE) )
 			syntax_error = 1;
 		else
 		{
 			
 			string_buffer_t mod;
 			sb_init( &mod );
-			wchar_t tmp[]=
-			{
-				BRACKET_END,
-				0
-			}
-			;
 			if( last_sep )
 			{
 				sb_append_substring( &mod, in, bracket_begin-in+1 );
@@ -978,7 +1051,9 @@ static int expand_brackets( wchar_t *in, int accept_incomplete, array_list_t *ou
 	
 	if( syntax_error )
 	{
-		error( SYNTAX_ERROR, L"Mismatched brackets", -1 );
+		error( SYNTAX_ERROR, 
+			   -1,
+			   L"Mismatched brackets" );
 		return 0;
 	}
 	
@@ -1006,7 +1081,7 @@ static int expand_brackets( wchar_t *in, int accept_incomplete, array_list_t *ou
 				wcsncpy( whole_item+len1, item_begin, item_len );	
 				wcscpy( whole_item+len1+item_len, bracket_end+1 );
 				
-				expand_brackets( whole_item, accept_incomplete, out );
+				expand_brackets( whole_item, flags, out );
 				
 				item_begin = pos+1;
 				if( pos == bracket_end )
@@ -1047,7 +1122,7 @@ int expand_locate_subshell( wchar_t *in,
 			if( wcschr( L"\'\"", *pos ) )
 			{
 				wchar_t *end = quote_end( pos );
-				if( end )
+				if( end && *end)
 				{
 					pos=end;
 				}
@@ -1063,8 +1138,7 @@ int expand_locate_subshell( wchar_t *in,
 				
 					paran_count++;
 				}
-			
-				if( *pos == ')' )
+				else if( *pos == ')' )
 				{
 					paran_count--;
 					if( (paran_count == 0) && (paran_end == 0) )
@@ -1102,6 +1176,11 @@ int expand_locate_subshell( wchar_t *in,
 	*begin = paran_begin;
 	*end = paran_count?in+wcslen(in):paran_end;
 	
+/*	assert( *begin >= in );
+	assert( *begin < (in+wcslen(in) ) );
+	assert( *end >= *begin );
+	assert( *end < (in+wcslen(in) ) );
+*/
 	return 1;
 
 }
@@ -1125,7 +1204,9 @@ static int expand_subshell( wchar_t *in, array_list_t *out )
 								   0 ) )
 	{
 		case -1:
-			error( SYNTAX_ERROR, L"Mismatched parans", -1 );
+			error( SYNTAX_ERROR, 
+				   -1,
+				   L"Mismatched parans" );
 			return 0;
 		case 0:
 			al_push( out, in );
@@ -1134,9 +1215,7 @@ static int expand_subshell( wchar_t *in, array_list_t *out )
 			
 			break;
 		
-	}
-	
-	
+	}	
 
 	len1 = (paran_begin-in);
 	len2 = wcslen(paran_end)-1;
@@ -1153,41 +1232,42 @@ static int expand_subshell( wchar_t *in, array_list_t *out )
 	wcsncpy( subcmd, paran_begin+1, paran_end-paran_begin-1 );
 	subcmd[ paran_end-paran_begin-1]=0;	
 	
-	if( !exec_subshell( subcmd, &sub_res) )
+	if( exec_subshell( subcmd, &sub_res)==-1 )
 	{
 		al_foreach( &sub_res, (void (*)(const void *))&free );
 		al_destroy( &sub_res );
 		free( subcmd );
 		return 0;
 	}	
-
+	
 	al_init( &tail_expand );
 	expand_subshell( wcsdup(paran_end+1), &tail_expand );
-
+	
     for( i=0; i<al_get_count( &sub_res ); i++ )
     {
-        wchar_t *sub_item;
+        wchar_t *sub_item, *sub_item2;
         sub_item = (wchar_t *)al_get( &sub_res, i );
-        sub_item = expand_escape( sub_item, 1 );
-        int item_len = wcslen( sub_item );
-
+        sub_item2 = expand_escape( sub_item, 1 );
+		free(sub_item);		
+        int item_len = wcslen( sub_item2 );
+		
         for( j=0; j<al_get_count( &tail_expand ); j++ )
         {
             string_buffer_t whole_item;
             wchar_t *tail_item = (wchar_t *)al_get( &tail_expand, j );
-
+			
             sb_init( &whole_item );
-
+			
             sb_append_substring( &whole_item, in, len1 );
             sb_append_char( &whole_item, INTERNAL_SEPARATOR );
-            sb_append_substring( &whole_item, sub_item, item_len );
+            sb_append_substring( &whole_item, sub_item2, item_len );
 			sb_append_char( &whole_item, INTERNAL_SEPARATOR );
             sb_append( &whole_item, tail_item );			
-
+			
             al_push( out, whole_item.buff );			
         }
 		
-        free( sub_item );
+        free( sub_item2 );
     }	
 	free(in);
 	
@@ -1195,195 +1275,47 @@ static int expand_subshell( wchar_t *in, array_list_t *out )
 	
 	al_foreach( &tail_expand, (void (*)(const void *))&free );
 	al_destroy( &tail_expand );
-
+	
 	free( subcmd );
 	return 1;	
 }
 
 
-wchar_t *expand_backslash( wchar_t * in, int escape_special )
+wchar_t *expand_unescape( const wchar_t * in, int escape_special )
 {
-	int in_pos, out_pos, len = wcslen( in );
-	int c;
-	int bracket_count=0;
-	wchar_t prev=0;	
-
-	for( in_pos=0, out_pos=0; in_pos<len; prev=in[out_pos], out_pos++, in_pos++ )
-	{
-		c = in[in_pos];
-		if( c == L'\\' )
-		{
-			switch( in[++in_pos] )
-			{
-				case L'\0':
-					error( SYNTAX_ERROR, L"Unexpected end of string", -1 );
-					return in;
-					
-					break;
-
-				case L'n':
-					in[out_pos]=L'\n';
-					break;
-
-				case L'r':
-					in[out_pos]=L'\r';
-					break;
-
-				case L't':
-					in[out_pos]=L'\t';
-					break;
-
-				case L'b':
-					in[out_pos]=L'\b';
-					break;
-
-				case L'e':
-					in[out_pos]=L'\e';
-					break;
-
-				default:
-					in[out_pos]=in[in_pos];
-					break;
-			}
-		}
-		else 
-		{
-			switch( in[in_pos]){
-				case L'~':
-					if( escape_special && (in_pos == 0) )
-						in[out_pos]=HOME_DIRECTORY;
-					else
-						in[out_pos] = L'~';
-					break;
-				case L'%':
-					if( escape_special && (in_pos == 0) )
-						in[out_pos]=PROCESS_EXPAND;
-					else
-						in[out_pos]=in[in_pos];						
-					break;
-				case L'*':
-					if( escape_special )
-						in[out_pos]=ANY_STRING;
-					else
-						in[out_pos]=in[in_pos];						
-					break;
-				case L'?':
-					if( escape_special )
-						in[out_pos]=ANY_CHAR;
-					else
-						in[out_pos]=in[in_pos];						
-					break;					
-				case L'$':
-					if( escape_special )
-						in[out_pos]=VARIABLE_EXPAND;
-					else
-						in[out_pos]=in[in_pos];											
-					break;					
-				case L'{':
-					if( escape_special )
-					{
-						bracket_count++;
-						in[out_pos]=BRACKET_BEGIN;
-					}
-					else
-						in[out_pos]=in[in_pos];						
-					break;					
-				case L'}':
-					if( escape_special )
-					{
-						bracket_count--;
-						in[out_pos]=BRACKET_END;
-					}
-					else
-						in[out_pos]=in[in_pos];						
-					break;					
-
-				case L',':
-					if( escape_special && bracket_count && prev!=BRACKET_SEP)
-					{
-						in[out_pos]=BRACKET_SEP;
-					}
-					else
-						in[out_pos]=in[in_pos];						
-					break;					
-
-				case L'\'':
-				case L'\"':
-				{
-					wchar_t *end = quote_end( &in[in_pos] );
-					int len;
-					
-					if( end == 0 )
-					{
-						error( SYNTAX_ERROR, L"Unexpected end of string", -1 );
-						return in;
-					}
-					
-					len = end- &in[in_pos]-1;					
-
-					if( escape_special)
-						in[out_pos++]=INTERNAL_SEPARATOR;
-
-					memmove( &in[out_pos], &in[in_pos+1], sizeof(wchar_t)*(len) );
-					in_pos += len+1;
-					out_pos += len-1;
-
-					if( escape_special)
-						in[++out_pos]=INTERNAL_SEPARATOR;
-
-					break;
-				}
-				default:
-					in[out_pos] = in[in_pos];
-					break;
-			}
-		}		
-	}
-	in[out_pos]=L'\0';
-	return in;	
+	wchar_t *res = unescape( in, escape_special );
+	if( !res )
+		error( SYNTAX_ERROR, -1, L"Unexpected end of string" );
+	return res;
 }
 
 /**
    Attempts tilde expansion. Of the string specified. If tilde
-   expansion is performed, the argument is freed and a new string is
-   allocated in its place. Horrible call signature. Should be
-   altered. Fugly!
+   expansion is performed, the original string is freed and a new
+   string allocated using malloc is returned, otherwise, the original
+   string is returned.
 */
-static int tilde_expand( wchar_t **ptr )
+static wchar_t * expand_tilde_internal( wchar_t *in )
 {
-	wchar_t *in = *ptr;
 	
 	if( in[0] == HOME_DIRECTORY )
 	{
 		int tilde_error = 0;
 		wchar_t *home=0;
-		wchar_t *new_in;
-		wchar_t *old_in;
+		wchar_t *new_in=0;
+		wchar_t *old_in=0;
 		
 //		fwprintf( stderr, L"Tilde expand ~%ls\n", (*ptr)+1 );
 		if( in[1] == '/' || in[1] == '\0' )
 		{
 			/* Current users home directory */
-			struct passwd *userinfo;
 			
-			while( 1 )
-			{
-				userinfo=getpwuid( getuid() );
-				if( userinfo )
-					break;
-				else if( errno != EINTR )
-				{
-					wperror( L"getpwuid" );
-				}				
-			}
-
-			if( userinfo )
-			{	
-				home = str2wcs(userinfo->pw_dir);
-			}
+			home = env_get( L"HOME" );
+			if( home )
+				home = wcsdup(home);
 			else
 				home = wcsdup(L"");
-
+			
 			if( !home )
 			{
 				*in = L'~';
@@ -1405,7 +1337,7 @@ static int tilde_expand( wchar_t **ptr )
 			old_in = name_end;
 			
 			char *name_str = wcs2str( name );			
-
+			
 			struct passwd *userinfo = 
 				getpwnam( name_str );
 			
@@ -1428,26 +1360,23 @@ static int tilde_expand( wchar_t **ptr )
 			free(name);
 		}
 		
-		if( !tilde_error )
+		if( !tilde_error && home && old_in )
 		{
 			new_in = wcsdupcat( home, old_in ); 
-			free( in );
-			in = new_in;
-			free(home);
-			*ptr = in;
-		}		
-		
+		}
+		free(home);
+		free( in );
+		return new_in;			
 	}
-	return 1;
+	return in;
 } 
 
-wchar_t *expand_tilde(wchar_t *in)
+wchar_t *expand_tilde( wchar_t *in)
 {
 	if( in[0] == L'~' )
 	{
 		in[0] = HOME_DIRECTORY;
-		tilde_expand( &in );
-		return in;
+		return expand_tilde_internal( in );
 	}
 	return in;	
 }
@@ -1461,8 +1390,6 @@ static void remove_internal_separator( const void *s, int conv )
 	wchar_t *in = (wchar_t *)s;
 	wchar_t *out=in;
 
-//	int changed=0;
-
 	while( *in )
 	{
 		switch( *in )
@@ -1486,32 +1413,36 @@ static void remove_internal_separator( const void *s, int conv )
 		}
 	}
 	*out=0;
-/*	if( changed )
-	{
-		fwprintf( stderr, L" -> %ls\n", s );
-	}
-*/
 }
 
 
 /**
-   The real expansion function. All other expansion  functions are wrappers to this one.
+   The real expansion function. expand_one is just a wrapper around this one.
 */
-static int expand_string_internal( wchar_t *str,
-								   array_list_t *end_out, 
-								   int flag,
-								   int accept_incomplete )
+int expand_string( wchar_t *str,
+				   array_list_t *end_out, 
+				   int flags )
 {
 	array_list_t list1, list2;
 	array_list_t *in, *out;
 	
 	int i;
 	int subshell_ok = 1;
+	int res = EXPAND_OK;
 	
+//	debug( 1, L"Expand %ls", str );
+	
+
+	if( (!(flags & ACCEPT_INCOMPLETE)) && is_clean( str ) )
+	{
+		al_push( end_out, str );
+		return EXPAND_OK;
+	}
+
 	al_init( &list1 );
 	al_init( &list2 );
 
-	if( EXPAND_SKIP_SUBSHELL & flag )
+	if( EXPAND_SKIP_SUBSHELL & flags )
 	{
 		wchar_t *pos = str;
 		
@@ -1523,10 +1454,11 @@ static int expand_string_internal( wchar_t *str,
 			
 			if( (pos == str) || ( *(pos-1) != L'\\' ) )
 			{
-				error( SUBSHELL_ERROR, L"Subshells not allowed", -1 );
+				error( SUBSHELL_ERROR, -1, L"Subshells not allowed" );
+				free( str );
 				al_destroy( &list1 );
 				al_destroy( &list2 );
-				return 0;
+				return EXPAND_ERROR;
 			}
 			pos++;
 		}		
@@ -1540,7 +1472,7 @@ static int expand_string_internal( wchar_t *str,
 	if( !subshell_ok )
 	{
 		al_destroy( &list1 );
-		return 0;
+		return EXPAND_ERROR;
 	}
 	else
 	{
@@ -1549,10 +1481,17 @@ static int expand_string_internal( wchar_t *str,
 
 		for( i=0; i<al_get_count( in ); i++ )
 		{
-			wchar_t *next = expand_backslash((wchar_t *)al_get( in, i ), 
-													  1);
+			wchar_t *next;
+			
+			next = expand_unescape( (wchar_t *)al_get( in, i ), 
+									1);
 
-			if( EXPAND_SKIP_VARIABLES & flag )
+			free( (void *)al_get( in, i ) );
+			
+			if( !next )
+				continue;			
+			
+			if( EXPAND_SKIP_VARIABLES & flags )
 			{
 				wchar_t *tmp;
 				for( tmp=next; *tmp; tmp++ )
@@ -1567,7 +1506,7 @@ static int expand_string_internal( wchar_t *str,
 				{
 					al_destroy( in );
 					al_destroy( out );
-					return 0;	
+					return EXPAND_ERROR;	
 				}
 			}
 		}
@@ -1580,11 +1519,11 @@ static int expand_string_internal( wchar_t *str,
 		for( i=0; i<al_get_count( in ); i++ )
 		{
 			wchar_t *next = (wchar_t *)al_get( in, i );
-			if( !expand_brackets( next, accept_incomplete, out ))
+			if( !expand_brackets( next, flags, out ))
 			{
 				al_destroy( in );
 				al_destroy( out );
-				return 0;
+				return EXPAND_ERROR;
 			}
 		}
 		al_truncate( in, 0 );
@@ -1595,32 +1534,37 @@ static int expand_string_internal( wchar_t *str,
 		for( i=0; i<al_get_count( in ); i++ )
 		{
 			wchar_t *next = (wchar_t *)al_get( in, i );		
-			if(!tilde_expand( &next ))
+			if( !(next=expand_tilde_internal( next ) ) )
 			{
 				al_destroy( in );
 				al_destroy( out );
-				return 0;
+				return EXPAND_ERROR;
 			}
 		
-			if( accept_incomplete )
+			if( flags & ACCEPT_INCOMPLETE )
 			{
 				if( *next == PROCESS_EXPAND )
 				{
-					expand_pid( next, accept_incomplete, end_out );
+					/*
+					  If process expansion matches, we are not
+					  interested in other completions, so we
+					  short-circut and return
+					*/
+					expand_pid( next, flags, end_out );
 					al_destroy( in );
 					al_destroy( out );
-					return 1;
+					return EXPAND_OK;
 				}
 				else
 					al_push( out, next );
 			}
 			else
 			{
-				if( !expand_pid( next, accept_incomplete, out ) )
+				if( !expand_pid( next, flags, out ) )
 				{
 					al_destroy( in );
 					al_destroy( out );
-					return 0;
+					return EXPAND_ERROR;
 				}
 			}
 		}
@@ -1628,49 +1572,50 @@ static int expand_string_internal( wchar_t *str,
 		
 		in = &list2;
 		out = &list1;		
-
 		
 		for( i=0; i<al_get_count( in ); i++ )
 		{
 			wchar_t *next = (wchar_t *)al_get( in, i );
 			int wc_res;
 
-			remove_internal_separator( next, EXPAND_SKIP_WILDCARDS & flag );				
-			if( ( accept_incomplete || wildcard_has( next, 1 ) ) )
+			remove_internal_separator( next, EXPAND_SKIP_WILDCARDS & flags );				
+			if( ((flags & ACCEPT_INCOMPLETE) && (!(flags & EXPAND_SKIP_WILDCARDS))) || 
+				wildcard_has( next, 1 ) )
 			{
 				
 				if( next[0] == '/' )
 				{
-					wc_res = wildcard_expand( &next[1], L"/",accept_incomplete, out );
+					wc_res = wildcard_expand( &next[1], L"/",flags, out );
 				}
 				else
 				{
-					wc_res = wildcard_expand( next, L"", accept_incomplete, out );
+					wc_res = wildcard_expand( next, L"", flags, out );
 				}
 				free( next );
 				switch( wc_res )
 				{
 					case 0:
-						if( !accept_incomplete )
+						if( !(flags & ACCEPT_INCOMPLETE) )
 						{	
+							if( res == EXPAND_OK )
+								res = EXPAND_WILDCARD_NO_MATCH;
 							break;
 						}
+						
 					case 1:
+						res = EXPAND_WILDCARD_MATCH;
 						sort_list( out );
 						al_push_all( end_out, out );
 						al_truncate( out, 0 );						
 						break;
 						
-					default:
-						fwprintf( stderr, L"error\n" );
-						/*al_destroy( &list1 );*/
-						/*al_destroy( &list2 );*/
-						/*return 0;*/
 				}			
 			}
 			else
 			{
-				if( !accept_incomplete)
+				if( flags & ACCEPT_INCOMPLETE)
+					free( next );
+				else
 					al_push( end_out, next );
 			}		
 		}
@@ -1678,29 +1623,24 @@ static int expand_string_internal( wchar_t *str,
 		al_destroy( out );
 	}
 
-	return 1;
+	return res;
+	
 }
 
 
-
-int expand_string( wchar_t *str, 
-				   array_list_t *out,
-				   int flag )
-{
-	return expand_string_internal( str, out, flag, 0 );
-}
-
-wchar_t *expand_one( wchar_t *string, int flag )
+wchar_t *expand_one( wchar_t *string, int flags )
 {
 	array_list_t l;
 	int res;
 	wchar_t *one;
+
+	if( (!(flags & ACCEPT_INCOMPLETE)) &&  is_clean( string ) )
+		return string;	
 	
 	al_init( &l );
-	res = expand_string_internal( string, &l, flag, 0 );
+	res = expand_string( string, &l, flags );
 	if( !res )
 	{
-		free( string );
 		one = 0;
 	}
 	else
@@ -1715,15 +1655,9 @@ wchar_t *expand_one( wchar_t *string, int flag )
 			al_set( &l, 0, 0 );
 		}
 	}
-	
+
 	al_foreach( &l, (void(*)(const void *))&free );
 	al_destroy( &l );
 	return one;
 }
 
-int expand_file_completion( wchar_t *str,
-							array_list_t *out )
-{
-	return expand_string_internal( str, out, 0, 1 );
-}
-
