@@ -2,6 +2,11 @@
   Prototypes for history functions, part of the user interface.
 */
 
+#ifndef FISH_HISTORY_H
+#define FISH_HISTORY_H
+
+#include <wchar.h>
+
 /**
    Load history from file.
 */
@@ -48,6 +53,7 @@ void history_reset();
 */
 const wchar_t *history_next_match( const wchar_t *str);
 
+
 /**
    Set the current mode name for history. Each application that uses
    the history has it's own mode. This must be called prior to any use
@@ -62,3 +68,4 @@ void history_set_mode( wchar_t *name );
 */
 void history_sanity_check();
 
+#endif
