@@ -22,6 +22,7 @@
 #include "reader.h"
 #include "kill.h"
 #include "wutil.h"
+#include "translate.h"
 
 /**
    Status from earlier sanity checks
@@ -30,14 +31,14 @@ static int insane;
 
 void sanity_lose()
 {
-	fwprintf( stderr, L"Errors detected. Shutting down.\n" );
+	debug( 0, _(L"Errors detected, shutting down") );
 	insane = 1;
 }
 
 int sanity_check()
 {
 	if( !insane )
-		if( shell_is_interactive )
+		if( is_interactive )
 			history_sanity_check();
 	if( !insane )
 		reader_sanity_check();
@@ -51,23 +52,22 @@ int sanity_check()
 
 void validate_pointer( const void *ptr, const wchar_t *err, int null_ok )
 {
-
+	
 	/* 
 	   Test if the pointer data crosses a segment boundary. 
 	*/
-
+	
 	if( (0x00000003 & (int)ptr) != 0 )
 	{
-		fwprintf( stderr, L"fish: The pointer '\%ls\' is invalid\n", err );
+		debug( 0, _(L"The pointer '%ls' is invalid"), err );
 		sanity_lose();		
 	}
 	
 	if((!null_ok) && (ptr==0))
 	{
-		fwprintf( stderr, L"fish: The pointer '\%ls\' is null\n", err );
+		debug( 0, _(L"The pointer '%ls' is null"), err );
 		sanity_lose();		
-	}	
-
+	}
 }
 
 
